#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		     Network Request Dispatcher  (NETSEND)
 *
 *	Usage:  netsend <host> [-i] [-p port] [remote-queue]
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 */

#include <stdio.h>
#include <signal.h>
#include "queue.h"
#include "qreplys.h"
#include <sys/stat.h>
#include <pwd.h>

time_t	timenow;
struct	request r;
int	options;
int	ourpid;
char	reqtmp[] = "/tmp/reqtmpXXXXXX";
FILE	*netfp;
char	*workdir = 0;
char	*user;
char	*queue;
char	*remotehost;
int	interactive = 0;
int	mdqsport = 666;

#define	SMALLWAIT	120	/* 2 minutes */
#define	BIGWAIT		900	/* 15 minutes */

char	*index();
char	*newtmp();
char	*strdup();
int	doflush();
int	dorestart();
FILE	*netopen();
int	timeout();
struct	request *gethdr();
struct	passwd *getpwuid();

extern	char *Qdatadir;

/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	linebuf[LINESIZ];
	char	newline[LINESIZ];
	char	datafile[PATHSIZ];
	char	*cp;
	int	fd;
	FILE	*tmpfp;
	struct stat sb;
	struct passwd *pw;
	int	i;
int oldformat = 0;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	time (&timenow);
	ourpid = getpid();
	
	if (argc < 2) {
		err ("MDQS NETSEND: Arg count wrong (argc = %d)", argc);
		cleanup(1);
		retmsg (RP_FATAL, "Incorrect arg count to netsend");
		exit (11);
	}
	
	for (i = 2; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			case 'i':
				interactive++;
				break;
			case 'o':
				oldformat++;
				break;
			case 'p':
				i++;
				mdqsport = atoi(argv[i]);
				break;
			default:
				retmsg (RP_FATAL, "Bad switch arg to netsend");
			}
		} else
			queue = argv[i];
	}
	if ((netfp = netopen (remotehost = argv[1], mdqsport)) == NULL) {
		retmsg (RP_DELAY, "Unable to connect to host %s", argv[1]);
		exit (13);
	}

	signal (SIGALRM, timeout);
	statmsg ("connection open");
	mktemp (reqtmp);
	if ((fd = creat(reqtmp, 0600)) < 0) {
		err ("MDQS NETSEND: %s: %s\n", reqtmp, sys_errlist[errno]);
		cleanup(1);
		retmsg (RP_FATAL, "Unable to create tempfile '%s'", reqtmp);
		exit (12);
	}
	close (fd);
	if ((tmpfp = fopen (reqtmp, "r+")) == NULL) {
		err ("MDQS NETSEND: Can't fdopen reqtmp\n");
		unlink (reqtmp);
		cleanup(1);
		retmsg (RP_FATAL, "Unable to open tempfile '%s'", reqtmp);
		exit (12);
	}
	unlink (reqtmp);
	
	if (gethdr (stdin, &r) == (struct request *) NULL) {
		err ("MDQS NETSEND: Error reading request\n");
		cleanup(1);
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}
	if (queue != 0)
		strncpy (r.r_que, queue, QUESIZ);
	r.r_tstart = 0L;
	puthdr (&r, tmpfp);
	if ((pw = getpwuid(getuid())) == NULL)
		user = "UnknownUser";
	else
		user = pw->pw_name;
	
	alarm (SMALLWAIT);
	fprintf (netfp, "Site: %s\n", "UnknownHost");	/* Superfluous */
	readreply (101, RP_FATAL);
if (!oldformat) {					/*** KLUDGE ***/
	alarm (SMALLWAIT);
	fprintf (netfp, "User: %s\n", user);
	readreply (102, RP_FATAL);
	alarm (SMALLWAIT);
	fprintf (netfp, "Queue: %.*s\n", QUESIZ, r.r_que);
	readreply (103, RP_FATAL);
}
	alarm (SMALLWAIT);
	fprintf (netfp, "Request: %.*s\n", QDIRSIZ, r.r_name);
	readreply (104, RP_FATAL);
	
	statmsg ("initial data sent");
	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		int	filecnt;
		char	msgbuf[LINESIZ];

		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'B':
			strcpy (newline, newtmp());
			alarm (SMALLWAIT);
			fprintf (netfp, "File: %s\n", newline);
			readreply (105, RP_DELAY);
			sprintf (datafile, "%s%s", Qdatadir, &linebuf[1]);
			sprintf (msgbuf, "sending %s", &linebuf[1]);
			statmsg (msgbuf);
			dumpfile (datafile);
			fprintf (tmpfp, "B%s\n", newline);
			break;

		case 'D':		/* Do NOT copy this line */
			if (workdir)
				free(workdir);
			workdir = strdup(&linebuf[1]);
			break;

		case 'F':
			strcpy (newline, newtmp());
			alarm (SMALLWAIT);
			fprintf (netfp, "File: %s\n", newline);
			readreply (105, RP_DELAY);
			sprintf (msgbuf, "sending %s", &linebuf[1]);
			statmsg (msgbuf);
			if( linebuf[1] == '/' ) {
				/* Absolute path */
				dumpfile (&linebuf[1]);
			} else {
				sprintf (datafile, "%s/%s", Qdatadir,
							&linebuf[1]);
				dumpfile (datafile);
			}
			fprintf (tmpfp, "I%s\n", newline);
			break;

		case 'H':
			strcpy (newline, newtmp());
			alarm (SMALLWAIT);
			fprintf (netfp, "File: %s\n", newline);
			readreply (105, RP_DELAY);
			sprintf (msgbuf, "sending %s", &linebuf[1]);
			statmsg (msgbuf);
			if( linebuf[1] == '/' ) {
				/* Absolute path */
				dumpfile (&linebuf[1]);
			} else {
				sprintf (datafile, "%s/%s", Qdatadir,
							&linebuf[1]);
				dumpfile (datafile);
			}
			fprintf (tmpfp, "B%s\n", newline);
			break;

		case 'I':
			strcpy (newline, newtmp());
			alarm (SMALLWAIT);
			fprintf (netfp, "File: %s\n", newline);
			readreply (105, RP_DELAY);
			sprintf (datafile, "%s%s", Qdatadir, &linebuf[1]);
			sprintf (msgbuf, "sending %s", &linebuf[1]);
			statmsg (msgbuf);
			dumpfile (datafile);
			fprintf (tmpfp, "I%s\n", newline);
			break;

		case 'O':
			options = atoi (&linebuf[1]);
			/* Fall through */

		default:
			fputs (linebuf, tmpfp);
			putc ('\n', tmpfp);
			break;
		}
	}

	statmsg ("finishing up");
	fflush (tmpfp);
	rewind (tmpfp);
	fstat (fileno(tmpfp), &sb);
	alarm (SMALLWAIT);
	fprintf (netfp, "File: %s\nSize: %ld\n", cp = newtmp(), sb.st_size);
	if (copycnt (tmpfp, netfp, cp, sb.st_size) != 0) {
		err ("MDQS NETSEND: phase error on reqtmp\n");
		retmsg (RP_RESTART, "Phase error on tempfile '%s'", reqtmp);
		exit (30);
	}
	fclose (tmpfp);
	fputs ("Done\n", netfp);
	fflush (netfp);
	statmsg ("request transfered");

	/* This read always gets done. */
	alarm (BIGWAIT);
	fseek (netfp, 0L, 1);	/* IMPORTANT, flips from writing to reading */
	if (fgets (linebuf, LINESIZ, netfp) == NULL)
		retmsg (RP_DELAY, "No final reply from host");
	else { 
		if (cp = index (linebuf, '\n'))
			*cp = 0;
		if (atoi(linebuf) != 250)
			retmsg (RP_DELAY, "Host %s returned: %s",
				remotehost, linebuf);
		else {
			retmsg (RP_OK, "%s:%s", remotehost, &linebuf[3]);
			cleanup(0);
		}
	}
	fprintf (netfp, "Bye %d\n", atoi(linebuf));
	fclose (netfp);
	exit (0);
}

/*
 *		D U M P F I L E ( )
 *  Dump a file onto the network with headers and trailers
 */
dumpfile( file )
char *file;
{
	FILE	*df;
	struct stat sb;

	if ((df = fopen (file, "r")) == NULL) {
		err ("MDQS NETSEND: Can't open %s\n", file);
		cleanup(1);
		retmsg (RP_FATAL, "Cannot open data file '%s'", file);
		exit (14);
	}
	fstat (fileno(df), &sb);
	alarm (SMALLWAIT);
	fprintf (netfp, "Size: %ld\n", sb.st_size);
	readreply (106, RP_DELAY);
	alarm (BIGWAIT);
	if (copycnt (df, netfp, file, sb.st_size) != 0) {
		err ("MDQS NETSEND: Premature EOF on %s (%s)\n",
			file, sys_errlist[errno]);
		cleanup(1);
		retmsg (RP_FATAL, "Premature EOF on data file '%s'", file);
		exit (15);
	}
	fclose (df);
	readreply (107, RP_DELAY);
	alarm (0);
}

char *
newtmp ()
{
	static char	file[16];

	sprintf (file, "%lx%x", timenow++, ourpid);
	return (file);
}

cleanup (preserve)
int	preserve;
{
	char	datafile[PATHSIZ];
	char	linebuf[LINESIZ];
	struct	request r;
	register char 	*cp;

	/* Remove data files */
	rewind (stdin);
	gethdr (stdin, &r);
	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'B':
		case 'I':
			sprintf (datafile, "%s%s", Qdatadir, &linebuf[1]);
			unlink (datafile);
			break;
		case 'F':
			if (!preserve && (options & NOKEEP))
				unlink (&linebuf[1]);
			break;
		}
	}
}

readreply (expect, failcode)
int	expect;
int	failcode;
{
	char	linebuf[LINESIZ];

	if (interactive) {
		fseek (netfp, 0L, 1);	/* IMPORTANT! Flip Write/Read mode */
		if (fgets (linebuf, LINESIZ, netfp) == NULL) {
			retmsg (RP_DELAY, "No reply from remote host");
			exit (0);
		}
		else if (atoi(linebuf) != expect) {
			retmsg (failcode, "Host %s returned: %s",
				remotehost, linebuf);
			if (failcode == RP_FATAL)
				cleanup(1);
			exit (0);
		}
		fseek (netfp, 0L, 10);	/* Flip Back. */
	}
}

timeout()
{
	alarm(0);
	err ("MDQS NETSEND: network timeout\n");
	retmsg (RP_RESTART, "network read or write timed out");
	exit (0);
}
